function [p_hat,gam_hat] = ps_MLE(X,D,ps_model)
% Estimate propensity score p(X,g) parameter g by MLE
% Requires a model: 1 = Laplace or 2 = Normal Fu(c) = P(U < c)
% ps(x,g) = Fu(g'x)
  
    num_Xs = size(X,2);

    if num_Xs == 1;                 % b on X
        A_const = [-1;1];
        b_const = [0,4];
    elseif num_Xs == 2;             % (a,b) on [1,X]
        A_const = [-1 0;
                    1 0;   
                    0 -1;
                    0 1];
        b_const = [0,1,0,4];
    elseif num_Xs == 3              % (g1,b,g2) on [Bernoulli,X,X^p]
        A_const = [-1 0 0;
                    1 0 0;
                    0 -1 0;
                    0 1 0;
                    0 0 -1;
                    0 0 1];
        b_const = [0,1,0,4,0,1];
    elseif num_Xs == 4              % (a,g1,b,g2) on [1,Bernoulli,X,X^p]
        A_const = [-1 0 0 0;
                    1 0 0 0;   
                    0 -1 0 0;
                    0 1 0 0;
                    0 0 -1 0;
                    0 0 1 0;
                    0 0 0 -1;
                    0 0 0 1];
        b_const = [0,1,0,1,0,4,0,1];
    end

    optimization_options_fmincon = optimoptions('fmincon', 'Algorithm', 'active-set', 'Diagnostics', 'off',...
                                           'TolX',1e-7,'TolFun',1e-7, 'MaxFunEvals', 1000, 'GradObj','on',...
                                          'MaxFunEvals', 20000, 'maxiter', 20000, 'DerivativeCheck', 'off', 'Display', 'off');
    %optimization_options_fmincon = optimoptions('fmincon', 'Algorithm', 'interior-point');
    
    gam0 = rand(size(X,2),1);
    [gam_hat,f_hat] =  fmincon(@ps_MEL_crit,gam0,A_const,b_const,[],[],[],[],[],optimization_options_fmincon);
    
    p_hat = ps_compute(gam_hat);

    return;

function [q,grad] = ps_MEL_crit(gam)
% MLE criterion for computing g in p(X,g)

    ps = ps_compute(gam);
    q = -mean(log((ps.^D).*((1-ps).^(1-D))));
    grad = grad_compute(gam);

end

function [ps] = ps_compute(gam)
% Compute propensity score p(x,g) based on model

    if ps_model == 1 % Normal Fu(c) = P(U < c)
        ps = normcdf(X*gam);
    end
    if ps_model == 2 % Laplace Fu(c) = P(U < c)
        ps = (.5*exp(X*gam)).*(X*gam <= 0) + (.5*(2-exp(-(X*gam)))).*(X*gam > 0);
    end
    
end

function [grad] = grad_compute(gam)
% Compute MLE gradient

    ps = ps_compute(gam);

    if ps_model == 1 % Normal Fu(c) = P(U < c)
        Xg = X*gam;
        Xg = repmat(Xg,1,size(X,2));
        dps = (1/(sqrt(2*pi)))*exp(-((Xg).^2)./2).*X;
    end
    if ps_model == 2 % Laplace Fu(c) = P(U < c)
        Xg = X*gam;
        IXg = (Xg <= 0);
        Xg = repmat(Xg,1,size(X,2));
        IXg = repmat(IXg,1,size(X,2));
        dps = (.5*exp(Xg)).*X.*IXg + (.5*exp(-(Xg))).*X.*(1-IXg);
    end
    h = D./ps - (1-D)./(1-ps);
    h = repmat(h,1,size(dps,2));
    grad = -mean(h.*dps);
    ps01 = (min(ps) == 0)+(max(ps) == 1);
    if ps01 > 0
        grad = ones(size(dps,2),1)*100;
    end

end

end