% ***********************************************************
% Tail-Trimmed ATE estimation with bias correctoin             
%                                                              
% Bias corrected estimator is optimally selected to be         
% arbitrarily close to the untrimmed estimator.
%
% The true propensity score, or an estimator, can be used.
% A parametric model is used for the propensity score: 
% MLE is used to compute a plug-in.
%                                                              
% DGP: latent variable treatment assignment D = I(b*X-U>0)      
%      and potential outcome Yj is iid                         
%      Distributions: Normal, Laplace, Pareto                  
%                                                              
%      We allow a nonadditively separatble D                   
%                                                              
% Jonathan B. Hill (2024)                                      
% ************************************************************

%% Initial setup

% DIRECTORY: place "Scripts" where it can be found by this program.
clc

%% GLOBAL VARIABLES


%% PRINT FLAGS FOR SIMULATION AND ESTIMATION
print_iters = 0;        % 1 = print output from each sample
print_best = 1;         % 1 = output only for No-Trim, TZ-BC, TX-Adapt(kx) : the best performers

%% ESTIMATION SETTINGS
tail_a_sym = 1;         % 1 = assume tail exponents a1 = a2 for bias estimation

z_center_trim = 1;      % 1 = use z-ave(z) for the trimming indicator 
scale_n_k = 1;          % 1 = scale by 1/(n - k) 
a_tol = .001;           % we only keep mn's if at least one a_hat[mn] > 1+a_tol

med_bias = 1;           % 1 = use median for optimal bias computatin; 0 = use min
fit_to_untrimmed = 1;   % 1 = use final criterion: choose ate_tt or ate_tt_bc based on how close to ate

estimate_ps = 1;        % 1 = estimate propensity score; 0 = use true ps
estimate_constant = 1;  % 1 = include constant term in parametric score model for estimation

lam_k = .25;            % k = lam_k*f(n) for some function of n: ln(n) or n^(2/3)
iota = 10^(-10);        % used for tail-trimming fractile relative to tail fractile 

%%  SIMULATION CONFIGURATION
max_rep = 1000;
ns = [100];

mn_scale = 5;           % mn window = 1,...,mn_scale*mn

other_Xs = [1,2];    % covariate types beside what is described in rv_types : put in ascending order
                            % 0 = do not use other X's
                            % 1 = Bernoulli; 
                            % 2 = sqrt(|X|^pow) where X comes from rv_types, pow from below
                            % as many X's can be "2" as there are different pow's, below

prob_bern = .3;         % Bernoulli covariate prob.
pows = 2;               % covariate power

rv_types = [2];         % 1 = normal; 2 = laplace;
                        % 3 = e,x ~ Normal and U ~ Laplace; 
                        % 4 = e,x ~ Laplace and U ~ normal 

a = 0;                  % if constant used: a = constant
bs = [.25,1,2];         % latent variable parameter in D = I(gam'X - U > 0) where gam = [g0,b,b/2]
g0 = .5;

kn_log = 1;             % 1 = logarithmic kn; else n^(2/3) 

%% Simulation Iterations and Storage

for i_rv = 1:length(rv_types);

    rv_type = rv_types(i_rv);

    for i_b = 1:length(bs);

        b = bs(i_b);
        
        for i_x = 1:size(other_Xs,1)
            
            other_X = other_Xs(i_x,:);         % covariate types
            if min(other_X) == 0                    % number of total covariantes = 2 (constant,X)
                num_Xs = 2;
                gam = [a,b]';
            else                                    % number of total covariates = 4 (constant,Bernoulli,X, X^p)
                num_Xs = 2 + length(other_X);
                gam = [a,g0,b,b/4]';
            end

            for i_n = 1:length(ns);

                n = ns(i_n);

                if kn_log == 1                        % for trimming by Z 
                    kn = max(1,round(lam_k*(log(n))^(1-iota)));    
                else
                    kn = max(1,round(lam_k*(n^(1/3-iota))));
                end

                knx = max(1,round(2*n/log(n)));        % for trimming by X: 2*n/log(n) 

                if kn_log == 1                          % for bias correction: tail exponent estimation 
                    mn = max(1,round(mn_scale*lam_k*log(n)));
                else
                    mn = max(1,round(lam_k*(n^(1/3))));  
                end

                %gn = .75*sqrt(log(n));                   % for Khan/Tamer 
                gn = log(log(n));

                ate = zeros(max_rep,1);
                ate_tt = zeros(max_rep,1);
                ate_tt_bc = zeros(max_rep,1);
                ate_x = zeros(max_rep,1);
                ate_x_ad = zeros(max_rep,1);
                ate_x_ad2 = zeros(max_rep,1);

                trim_tt  = zeros(max_rep,1);
                trim_tt_bc  = zeros(max_rep,1);
                trim_x  = zeros(max_rep,1);
                trim_x_ad  = zeros(max_rep,1);
                trim_x_ad2  = zeros(max_rep,1);
                
                gam_hat = zeros(max_rep,length(gam)+(estimate_constant-1));

                    % sample iterations
                for r = 1:max_rep

                        % redraw sample if a problem arises
                    bad_sample = 1;
                    while bad_sample == 1;
                        bad_sample = 0;         % flip to 1 and repeat sample draw if one of several problems arise 

                        if rv_type == 1         % e,x,u ~ normal 
                            zs = randn(n,4);
                        end
                        if rv_type == 2;        % e,x,u ~ Laplace 
                            zs = laplace_gen(n,4);
                        end
                        if rv_type == 3;        % e,x ~ normal, u ~ Laplace 
                            z0 = randn(n,3);
                            z1 = laplace_gen(n,1);
                            zs = cat(2,z0,z1);
                        end
                        if rv_type == 4;        % e,x ~ Laplace, u ~ normal 
                            z0 = laplace_gen(n,3);
                            z1 = randn(n,1);
                            zs = cat(2,z0,z1);
                        end

                        e0 = zs(:,1);
                        e1 = zs(:,2);
                        X = zs(:,3);
                        U = zs(:,4);
                        
                            % buiuld all covariates Xs using X and its powers
                        Xs = zeros(n,num_Xs);
                        Xs(:,1) = ones(length(X),1);
                        Xs(:,2) = X;                    % at least two covariates: (constant,X)
                        pow_count = 1;
                        if num_Xs > 2                   % add other covariates
                            for j = 1:length(other_X)
                               if other_X(j) == 1;
                                   u = rand(n,1);
                                   Xs(:,2+j) = u.*(u <= prob_bern);
                               end
                               if other_X(j) == 2;
                                   Xs(:,2+j) = abs(X).^pows(pow_count);
                                   pow_count = pow_count+1;
                               end   
                            end
                        end
                        
                        if estimate_constant == 0
                            Xs_estim = Xs(:,2:size(Xs,2));
                        else
                            Xs_estim = Xs;
                        end
                        num_Xs_estim = size(Xs_estim,2);

                        D = (Xs*gam-U > 0);
            
                        if rv_type == 1         % e,X,U ~ normal 
                            pX = normcdf(Xs*gam);
                            ps_model = 1;
                        end 
                        if rv_type == 2       % e,X,U ~ Laplace 
                            pX = (.5*exp(Xs*gam)).*(Xs*gam <= 0) + (.5*(2-exp(-(Xs*gam)))).*(Xs*gam > 0);
                            ps_model = 2;
                        end
                        if rv_type == 3       % e,X ~ normal, U ~ Laplce 
                            pX = (.5*exp(Xs*gam)).*(Xs*gam <= 0) + (.5*(2-exp(-(Xs*gam)))).*(Xs*gam > 0);
                            ps_model = 2;
                        end
                        if rv_type == 4       % e,X ~ Laplace, U ~ normal 
                            pX = normcdf(Xs*gam);
                            ps_model = 1;
                        end

                        if estimate_ps == 1;
                            [pX_hat,gam_hat(r,:)] = ps_MLE(Xs_estim,D,ps_model);
                        else
                            px_hat = pX;
                        end

                        Y0 = e0;
                        Y1 = e1;
                        Y = D.*e1+(1-D).*e0;
                        h = D./pX_hat + (1-D)./(1-pX_hat);

                        Z = Y.*h;

                        if isnan(Z) == 1
                            if print_iters == 1
                                disp('Missing values: redraw data.');
                            end
                            bad_sample = 1;
                            continue;
                        end

                        if z_center_trim == 1
                            ZZ = Z - mean(Z);
                        else
                            ZZ = Z;
                        end

                        Za = sort(abs(ZZ),1);
                        cz = Za(length(Za)-kn+1,1);
                        Xa = sort(abs(X),1);
                        cx = Xa(length(Xa)-knx+1,1);
                        cx2 = Xa(length(Xa)-kn+1,1);

                            % mn's and tail exponents in range of mn's 
                        mns = 1:min(cat(2,20*mn,.75*sum(ZZ >= 0),.75*sum(ZZ < 0)));
                        [azms,czms,dzms] = a_hat(ZZ,mns);
                        [a1zms,c1zms,d1zms] = a_hat(ZZ.*(ZZ < 0),mns);
                        [a2zms,c2zms,d2zms] = a_hat(ZZ.*(ZZ >= 0),mns);

                        if isnan(azms)+isnan(a1zms)+isnan(a2zms) > 0
                            if print_iters == 1
                                disp('missing a_hats: redraw sample');
                            end
                            bad_sample = 1;
                            continue;
                        end

                        ate(r,1) = mean(Z);

                        if z_center_trim == 1
                            ate_tt(r,1) = sum(Z.*(abs(ZZ) < cz))/(length(Z)-kn);
                        else
                            ate_tt(r,1) = mean(Z.*(abs(ZZ) < cz));
                        end

                        ate_x(r,1) = mean(Z.*(abs(X) < sqrt(log(n))));
                        ate_x_ad(r,1) = mean(Z.*(abs(X) < cx));         % cx based on knx
                        ate_x_ad2(r,1) = mean(Z.*(abs(X) < cx2));       % cx2 based on kn < knx

                        trim_tt(r,1) = mean(abs(ZZ) >= cz);
                        trim_tt_bc(r,1) = mean(abs(ZZ) >= cz);       % core bias corrected estimator: the uncorrected estimator 
                        trim_x(r,1) = mean(abs(X) >= gn);
                        trim_x_ad(r,1) = mean(abs(X) >= cx);
                        trim_x_ad2(r,1) = mean(abs(X) >= cx2);

                            % Compute optimal bias corrected ATE estimator 
                                    % remove mn's if a(mn) < 1+tol
                        if tail_a_sym == 1;
                            d = cat(2,mns',azms,d1zms,d2zms);
                        else
                            a1zms(a1zms < 1+a_tol) = NaN;
                            a2zms(a2zms < 1+a_tol) = NaN;
                            d = cat(2,mns',a1zms,a2zms,d1zms,d2zms);
                        end
                        d = d(all(~isnan(d),2),:);

                        if size(d,1)*size(d,2) == 0
                            if print_iters == 1
                                disp('a_hats too small: re-draw sample');
                            end
                            bad_sample = 1;
                            continue;
                        elseif isnan(d) == 1
                            if print_iters == 1
                                disp('a_hats too small: re-draw sample');
                            end
                            bad_sample = 1;
                            continue;
                        end

                        mns = d(:,1);
                        if tail_a_sym == 1;
                            azms = d(:,2);
                            d1zms = d(:,3);
                            d2zms = d(:,4);
                        else
                            a1zms = d(:,2);
                            a2zms = d(:,3);
                            d1zms = d(:,4);
                            d2zms = d(:,5);
                        end

                                % bias and bias corrected ATE's on range of mn's
                        if tail_a_sym == 0
                            bns = (n/(n-kn))*( (d2zms.^(1./a2zms)).*(a2zms./(a2zms-1)).*((kn/n).^(1-1./a2zms))...
                                               -(d1zms.^(1./a1zms)).*(a1zms./(a1zms-1)).*((kn/n).^(1-1./a1zms)) );
                        else
                            bns = (n/(n-kn))*((kn/n).^(1-1./azms)).*((azms./(azms-1))).*( (d2zms.^(1./azms))-(d1zms.^(1./azms)) );
                        end

                        bns = cat(1,bns-1/n,bns,bns+1/n);

                                % optimal bias-corrected estimator 
                        ate_tt_bcs = ate_tt(r,1)+bns;
                        d = cat(2,ate_tt_bcs,abs(ate_tt_bcs-ate(r,1)));
                        [~,indexd] = sort(d(:,2),'ascend');
                        d = d(indexd,:);                    % sort by 2'nd collumn
                        if med_bias == 1
                            ate_tt_bc_star = median(ate_tt_bcs);
                        else
                            ate_tt_bc_star = d(1,1);
                        end

                                % choose which ever estimator, tt_bc_star or tt, is closer to untrimmed
                        if fit_to_untrimmed == 1;
                            ate_tt_bc_starstar = ate_tt_bc_star*(abs(ate_tt_bc_star-ate(r,1)) < abs(ate_tt(r,1)-ate(r,1)))...
                                                + ate_tt(r,1)*(abs(ate_tt_bc_star-ate(r,1)) >= abs(ate_tt(r,1)-ate(r,1)));
                        else
                            ate_tt_bc_starstar = ate_tt_bc_star;
                        end
                        ate_tt_bc(r,1) = ate_tt_bc_starstar;

                        if print_iters == 1;
                            fprintf('rep %d out of %d. sample size n = %d\n',r,max_rep,n);
                            disp(' ');
                            disp('     ATE      ATE TT  ATE TT(bc)    KT   KT-adaptive');
                            disp(cat(2,mean(ate(1:r,1)),mean(ate_tt(1:r,1)),mean(ate_tt_bc(1:r),1),mean(ate_x(1:r,1)),mean(ate_x_ad(1:r),1)));
                            disp(' ');
                        end

                    end

                end

                ate = ate(all(~isnan(ate),2),:);
                ate_tt = ate_tt(all(~isnan(ate_tt),2),:);
                ate_tt_bc = ate_tt_bc(all(~isnan(ate_tt_bc),2),:);
                ate_x = ate_x(all(~isnan(ate_x),2),:);
                ate_x_ad = ate_x_ad(all(~isnan(ate_x_ad),2),:);

                if size(ate,1) < max_rep/2;continue;end
                if size(ate_tt,1) < max_rep/2;continue;end
                if size(ate_tt_bc,1) < max_rep/2;continue;end
                if size(ate_x,1) < max_rep/2;continue;end
                if size(ate_x_ad,1) < max_rep/2;continue;end

                z = (ate)./sqrt(mean((ate).^2));
                rej = mean(repmat(abs(z),1,3) > repmat(cat(2,2.57,1.96,1.645),length(z),1))';
                [ad_cv5,ks_cv5] = normal_test(z);

                z_tt = (ate_tt)./sqrt(mean((ate_tt).^2));
                rej_tt = mean(repmat(abs(z_tt),1,3) > repmat(cat(2,2.57,1.96,1.645),length(z_tt),1))';
                [ad_cv5_tt,ks_cv5_tt] = normal_test(z_tt);

                ate_tt_bc_clip = ate_tt_bc;
                    ate_tt_bc_clip = sort(ate_tt_bc);
                    ate_tt_bc_clip = ate_tt_bc_clip(max(2,.002*length(ate_tt_bc_clip)):min(length(ate_tt_bc_clip)-1,.998*length(ate_tt_bc_clip)));
                    ate_tt_bc_clip=ate_tt_bc_clip-mean(ate_tt_bc_clip);
                z_tt_bc = (ate_tt_bc_clip)./sqrt(mean((ate_tt_bc_clip).^2));
                rej_tt_bc = mean(repmat(abs(z_tt_bc),1,3) > repmat(cat(2,2.57,1.96,1.645),length(z_tt_bc),1))';
                [ad_cv5_tt_bc,ks_cv5_tt_bc] = normal_test(z_tt_bc);

                z_x = (ate_x)./sqrt(mean((ate_x).^2));
                rej_x = mean(repmat(abs(z_x),1,3) > repmat(cat(2,2.57,1.96,1.645),length(z_x),1))';
                [ad_cv5_x,ks_cv5_x] = normal_test(z_x);

                z_x_ad = (ate_x_ad)./sqrt(mean((ate_x_ad).^2));
                rej_x_ad = mean(repmat(abs(z_x_ad),1,3) > repmat(cat(2,2.57,1.96,1.645),length(z_x_ad),1))';
                [ad_cv5_x_ad,ks_cv5_x_ad] = normal_test(z_x_ad);

                z_x_ad2 = (ate_x_ad2)./sqrt(mean((ate_x_ad2).^2));
                rej_x_ad2 = mean(repmat(abs(z_x_ad2),1,3) > repmat(cat(2,2.57,1.96,1.645),length(z_x_ad2),1))';
                [ad_cv5_x_ad2,ks_cv5_x_ad2] = normal_test(z_x_ad2);

                mse_sq_ate = sqrt(mean((ate).^2));
                mse_sq_ate_tt = sqrt(mean((ate_tt).^2));
                mse_sq_ate_tt_bc = sqrt(mean((ate_tt_bc).^2));
                mse_sq_ate_x = sqrt(mean((ate_x).^2));
                mse_sq_ate_x_ad = sqrt(mean((ate_x_ad).^2));
                mse_sq_ate_x_ad2 = sqrt(mean((ate_x_ad2).^2));

                info_ate = cat(1,mean(ate), median(ate),mse_sq_ate);
                info_ate_tt = cat(1,mean(ate_tt), median(ate_tt),mse_sq_ate_tt);
                info_ate_tt_bc = cat(1,mean(ate_tt_bc), median(ate_tt_bc),mse_sq_ate_tt_bc);
                info_ate_x = cat(1,mean(ate_x), median(ate_x),mse_sq_ate_x);
                info_ate_x_ad = cat(1,mean(ate_x_ad), median(ate_x_ad),mse_sq_ate_x_ad);
                info_ate_x_ad2 = cat(1,mean(ate_x_ad2), median(ate_x_ad2),mse_sq_ate_x_ad2);

                info_ksad5 = cat(1,ad_cv5,ks_cv5);
                info_ksad5_tt = cat(1,ad_cv5_tt,ks_cv5_tt);
                info_ksad5_tt_bc = cat(1,ad_cv5_tt_bc,ks_cv5_tt_bc);
                info_ksad5_x = cat(1,ad_cv5_x,ks_cv5_x);
                info_ksad5_x_ad = cat(1,ad_cv5_x_ad,ks_cv5_x_ad);
                info_ksad5_x_ad2 = cat(1,ad_cv5_x_ad2,ks_cv5_x_ad2);

                disp(' ');
                disp('----------------------');
                if estimate_ps == 1
                    disp('Propensity score is estimated.');
                else
                    disp('True propensity score is used.');
                end
                
                if num_Xs== 2
                    disp('True Covariates = [1,X]');
                else
                    disp('True Covariates = [1,X(1)] and:');
                    pow_count = 1;
                    for j = 1:length(other_X)
                        if other_X(j) == 1;fprintf('    X(%d) = Bernoulli\n',j+2);end;
                        if other_X(j) == 2;fprintf('    X(%d) = X(1)^%d\n',j+2,pows(pow_count));pow_count=pow_count+1;end;
                    end
                end
               
                if estimate_ps == 1
                    if estimate_constant == 1
                        if num_Xs_estim == 2
                            disp('    Estimation Covariates = [1,X]');
                        else
                            disp('    Estimation Covariates = [1,X(1)] and:');
                            pow_count = 1;
                            for j = 1:length(other_X)
                                if other_X(j) == 1;fprintf('    X(%d) = Bernoulli\n',j+2);end;
                                if other_X(j) == 2;fprintf('    X(%d) = X(1)^%d\n',j+2,pows(pow_count));pow_count=pow_count+1;end;
                            end
                        end
                    else
                        if num_Xs_estim == 1
                            disp('Estimation Covariates = [X]');
                        else
                            disp('Estimation Covariates = [X(1)] and:');
                            disp('    [X(1)]');
                            pow_count = 1;
                            for j = 1:length(other_X)
                                if other_X(j) == 1;fprintf('    X(%d) = Bernoulli\n',j+1);end;
                                if other_X(j) == 2;fprintf('    X(%d) = X(1)^%d\n',j+1,pows(pow_count));pow_count=pow_count+1;end;
                            end
                        end
                    end
                end
                if tail_a_sym == 1;disp('Assume tail index symmetry: k1 = k2');
                else disp('Tail index asymmetry allows: k1 <=> k2');end
                if med_bias == 1;disp('median used for bias selection');
                else disp('minumum used for bias selection');end;
                if fit_to_untrimmed == 1;disp('choose ate_tt_opt in (ate_tt,ate_tt_bc) close to ate');
                else disp('choose ate_tt_bc');end;
                if kn_log == 1;disp('logarithmic kn');else disp('kn = n^(2/3)');end
                fprintf('number of samples = %d\n',max_rep);
                disp('----------------------');
                if rv_type == 1;disp('e,x,u ~ normal');end
                if rv_type == 2;disp('e,x,u ~ laplace');end
                if rv_type == 3;disp('e,x ~ normal, u ~ laplace');end
                if rv_type == 4;disp('e,x ~ laplace, u ~ normal');end
                fprintf('(a,b) = (%2.2f,%2.2f)\n',a,b);
                fprintf('n = %2.2f\n',n);
                fprintf('trimming kn = %2.2f kn/n = %2.2f\n',kn,kn/n);
                fprintf('trimming {kn_x1,kn_x2} = {%2.2f,%2.2f} {kn_x1/n,kn_x2/n} = {%2.2f,%2.2f}\n',knx,kn,knx/n,kn/n);
                fprintf('tail mn = %2.2f\n',mn);
                disp('----------------------');
                if print_best == 1;
                    disp('      ATE    T.T ATE(bc)   KT-ad');
                    disp(cat(2,info_ate,info_ate_tt_bc,info_ate_x_ad));
                    disp(cat(2,info_ksad5,info_ksad5_tt_bc,info_ksad5_x_ad));
                    disp(' ');
                    disp('      Rej    Rej.TT(bc)  Rej KTad');
                    disp(cat(2,rej,rej_tt_bc,rej_x_ad));
                    disp(' ');
                    disp('  Trimmed Observations Percentage')
                    disp('       No T   TT(bc)    KT-ad');
                    disp(cat(2,0,mean(trim_tt_bc),mean(trim_x_ad)));              
                else
                    disp('      ATE    T.T.ATE  T.T ATE(bc)   KT       KT-ad    KT-ad2');
                    disp(cat(2,info_ate,info_ate_tt,info_ate_tt_bc,info_ate_x,info_ate_x_ad,info_ate_x_ad2));
                    disp(cat(2,info_ksad5,info_ksad5_tt,info_ksad5_tt_bc,info_ksad5_x,info_ksad5_x_ad,info_ksad5_x_ad2));
                    disp(' ');
                    disp('      Rej    Rej.TT   Rej.TT(bc)  Rej KT   Rej KTad  Rej KTad2');
                    disp(cat(2,rej,rej_tt,rej_tt_bc,rej_x,rej_x_ad,rej_x_ad2));
                    disp(' ');
                    disp('  Trimmed Observations Percentage')
                    disp('     No T       TT      TT(bc)      KT       KT-ad    KT-ad2');
                    disp(cat(2,0,mean(trim_tt),mean(trim_tt_bc),mean(trim_x),mean(trim_x_ad),mean(trim_x_ad2)));
                end
                if num_Xs == 1
                    fprintf('gam0 = %2.4f\n',gam);
                    fprintf('gam^ = %2.4f\n',mean(gam_hat));
                elseif num_Xs == 2
                    fprintf('gam0 = %2.4f %2.4f\n',gam);
                    if estimate_constant == 1
                        fprintf('gam^ = %2.4f %2.4f\n',mean(gam_hat));
                    else
                        fprintf('gam^ = %2.4f\n',mean(gam_hat));
                    end
                elseif num_Xs == 3
                    fprintf('gam0 = %2.4f  %2.4f  %2.4f\n',gam);
                    if estimate_constant == 1
                        fprintf('gam^ = %2.4f %2.4f %2.4f\n',mean(gam_hat));
                    else
                        fprintf('gam^ = %2.4f %2.4f\n',mean(gam_hat));
                    end
                elseif num_Xs == 4
                    fprintf('gam0 = %2.4f  %2.4f  %2.4f %2.4f\n',gam);
                    if estimate_constant == 1
                        fprintf('gam^ = %2.4f %2.4f %2.4f %2.4f\n',mean(gam_hat));
                    else
                        fprintf('gam^ = %2.4f %2.4f %2.4f\n',mean(gam_hat));
                    end
                end
                disp('----------------------');
            end
        end
    end
end

disp(' ');disp(' ')
disp('******************')
disp('End of Program Run')
disp('******************')